#include "mbed.h"
#include "BLE.h"
#include "GattAttribute.h"
#include "DeviceInformationService.h"

// name of the device
const static uint8_t DEVICE_NAME[] = {'D','e','v','v','y','\0'};

const char myArr[]       = {0x3d, 0xda, 0x00, 0x01, 0x95, 0x7f, 0x7d, 0x4a, 0x34, 0xa6, 0x74, 0x69, 0x66, 0x73, 0x69, 0x6d,'\0'};
const char myArrOne[]    = {0x3d, 0xda, 0x00, 0x02, 0x95, 0x7f, 0x7d, 0x4a, 0x34, 0xa6, 0x74, 0x69, 0x66, 0x73, 0x69, 0x6d,'\0'};
const char myArrTwo[]    = {0x3d, 0xda, 0x00, 0x03, 0x95, 0x7f, 0x7d, 0x4a, 0x34, 0xa6, 0x74, 0x69, 0x66, 0x73, 0x69, 0x6d,'\0'};
const char myArrThree[]  = {0x3d, 0xda, 0x00, 0x04, 0x95, 0x7f, 0x7d, 0x4a, 0x34, 0xa6, 0x74, 0x69, 0x66, 0x73, 0x69, 0x6d,'\0'};
const char myArrFour[]   = {0x3d, 0xda, 0x00, 0x05, 0x95, 0x7f, 0x7d, 0x4a, 0x34, 0xa6, 0x74, 0x69, 0x66, 0x73, 0x69, 0x6d,'\0'};
const uint8_t address1[] = {0x30, 0x4E, 0x3C, 0x96, 0x47, 0xFB,'\0'};

char valueZero[]  = {0x00, 0x00, 0x01, 0x01, 0x4c, 0x46, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xd0, 0xeb, 0x48,'\0'};
char valueOne[]   = {0x01, 0x24, 0x56, 0x10, 0xff, 0x00, 0xeb, 0x48, 0x24, 0x56, 0xc6, 0x03, 0x10, 0xff, 0xa0, 0x82, 0x00, 0x00, 0x00, 0x00,'\0'};
char valueTwo[]   = {0x02, 0x00, 0x00, 0xec, 0x48, 0x24, 0x56, 0x42, 0x00, 0x10, 0xff, 0xa0, 0x83, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xec,'\0'};
char valueThree[] = {0x03, 0x48, 0x24, 0x56, 0x45, 0x00, 0x10, 0xff, 0xa0, 0x80, 0x00, 0xc4, 0x09, 0x00, 0x00, 0x00, 0xec, 0x48, 0x24, 0x56,'\0'};
char valueFour[]  = {0x04, 0xa6, 0x00, 0x10, 0xff, 0xa0, 0x81, 0x00, 0xc4, 0x09, 0x00, 0x00, 0x00, 0xec, 0x48, 0x24, 0x56, 0xa8, 0x00, 0x10,'\0'};
char valueFive[]  = {0x05, 0xff, 0x90, 0x81, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xed, 0x48, 0x24, 0x56, 0x1b, 0x02, 0x10, 0xff, 0x80, 0x80,'\0'};
char valueSix[]   = {0x06, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x41, 0x49, 0x24, 0x56, 0xdc, 0x03, 0x10, 0xff, 0x90, 0x80, 0x27, 0x00, 0x00,'\0'};
char valueSeven[] = {0x07, 0x00, 0xd0, 0x07, 0x61, 0x49, 0x24, 0x56, 0x88, 0x03, 0x10, 0xff, 0x88, 0x01, 0x01, 0x01, 0x00, 0x00, 0xff, 0xff,'\0'};
char valueEight[] = {0x08, 0x63, 0x49, 0x24, 0x56, 0x0a, 0x03, 0x10, 0xff, 0x80, 0x85, 0x01, 0x01, 0x01, 0x1c, 0x00, 0x01, 0x63, 0x49, 0x24,'\0'};
char valueNine[]  = {0x09, 0x56, 0x13, 0x03, 0x10, 0xff, 0x80, 0x81, 0x01, 0x00, 0x00, 0x01, 0x1c, 0x00, 0x63, 0x49, 0x24, 0x56, 0x6e, 0x03,'\0'};
char valueTen[]   = {0x8a, 0x10, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,'\0'};

// GATT service and characteristic UUIDs
const UUID nRF51_GATT_SERVICE     = UUID((uint8_t *)myArr);
const UUID nRF51_GATT_CHAR_ONE    = UUID((uint8_t *)myArrOne);
const UUID nRF51_GATT_CHAR_TWO    = UUID((uint8_t *)myArrTwo);
const UUID nRF51_GATT_CHAR_THREE  = UUID((uint8_t *)myArrThree);
const UUID nRF51_GATT_CHAR_FOUR   = UUID((uint8_t *)myArrFour);

#define CHARACTERISTIC_ONE    0
#define CHARACTERISTIC_TWO    1
#define CHARACTERISTIC_THREE  2
#define CHARACTERISTIC_FOUR   3
#define CHARACTERISTIC_COUNT  4
#define timerONE 65535
#define waiting 0.2

DigitalOut myled(LED1);
DigitalOut myledTwo(LED2);
DigitalOut myledThree(LED3);
DigitalOut myledFour(LED4);


// our bluetooth smart objects
BLEDevice           ble;
GattService        *gatt_service;
GattService        *gatt_serviceONE;
GattCharacteristic *gatt_characteristics[CHARACTERISTIC_COUNT];
GattCharacteristic *gattInfo_characteristics[CHARACTERISTIC_COUNT];
uint8_t             gatt_char_value[CHARACTERISTIC_COUNT];

Gap::ConnectionParams_t fast;

void onConnection(const Gap::ConnectionCallbackParams_t *params)
{

}

void errorCatch(ble_error_t didSendValue){
    int i = 0;
    if(didSendValue == BLE_ERROR_BUFFER_OVERFLOW){
        while(i < BLE_ERROR_BUFFER_OVERFLOW){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_NOT_IMPLEMENTED){
        while(i < BLE_ERROR_NOT_IMPLEMENTED){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_PARAM_OUT_OF_RANGE){
        while(i < BLE_ERROR_PARAM_OUT_OF_RANGE){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_INVALID_PARAM){
        while(i < BLE_ERROR_INVALID_PARAM){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_STACK_BUSY){
        while(i < BLE_STACK_BUSY){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_INVALID_STATE){
        while(i < BLE_ERROR_INVALID_STATE){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_NO_MEM){
        while(i < BLE_ERROR_NO_MEM){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_OPERATION_NOT_PERMITTED){
        while(i < BLE_ERROR_OPERATION_NOT_PERMITTED){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_INITIALIZATION_INCOMPLETE){
        while(i < BLE_ERROR_INITIALIZATION_INCOMPLETE){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_ALREADY_INITIALIZED){
        while(i < BLE_ERROR_ALREADY_INITIALIZED){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_UNSPECIFIED){
        while(i < BLE_ERROR_UNSPECIFIED){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }else if(didSendValue == BLE_ERROR_INTERNAL_STACK_FAILURE){
        while(i < BLE_ERROR_INTERNAL_STACK_FAILURE){
            myledThree = !myledThree;
            wait(waiting);
            myledThree = !myledThree;
            i++;
        }
    }           
}

int sender(int condStatus)
{
    
    ble_error_t didSendValue = BLE_ERROR_NONE;

    if(condStatus == 0) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueZero, sizeof(valueZero));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled; 
        condStatus++;
    }
    
    wait(waiting);
    
    if(condStatus == 1) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueOne, sizeof(valueOne));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);
    
    if(condStatus == 2) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueTwo, sizeof(valueTwo));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);

    if(condStatus == 3) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueThree, sizeof(valueThree));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);

    if(condStatus == 4) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueFour, sizeof(valueFour));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);

    if(condStatus == 5) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueFive, sizeof(valueFive));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);

    if(condStatus == 6) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueSix, sizeof(valueSix));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);

    if(condStatus == 7) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueSeven, sizeof(valueSeven));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);

    if(condStatus == 8) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueEight, sizeof(valueEight));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);

    if(condStatus == 9) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueNine, sizeof(valueNine));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);

    if(condStatus == 10) {
        didSendValue = ble.gattServer().write(gatt_characteristics[CHARACTERISTIC_THREE]->getValueHandle(), (uint8_t *)valueTen, sizeof(valueTen));
        if (didSendValue != BLE_ERROR_NONE) {
            myledTwo = !myledTwo;
            errorCatch(didSendValue);
            myledTwo = !myledTwo;
            return condStatus;
        }
        myled = !myled;
        wait(waiting);
        myled = !myled;
        condStatus++;
    }
    
    wait(waiting);
    
    return condStatus;
}


void onDataWritten(const GattWriteCallbackParams *context)
{
    // was the characteristic being written to nRF51_GATT_CHAR_LED?
    if (context -> handle == gatt_characteristics[CHARACTERISTIC_TWO] -> getValueHandle()) {

        int condStatus = 0;
        int loopCount = 0;

        while(condStatus <= 10 && loopCount <= 5) {
            condStatus = sender(condStatus);
            loopCount++;
            myled = 1;
            myledTwo = 1;
        }

    }
}

void onDisconnection(const Gap::DisconnectionCallbackParams_t *params)
{
    BLE& ble = BLE::Instance(BLE::DEFAULT_INSTANCE);
    ble.startAdvertising(); // restart advertising
}

int main()
{
    // wait a second before trying to write something to console
    myled = 1;
    myledTwo = 1;
    myledThree = 1;
    myledFour = 1;
    wait(1);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    // create our button characteristic (read, notify)
    gatt_characteristics[CHARACTERISTIC_ONE] =
        new GattCharacteristic(
        nRF51_GATT_CHAR_ONE,
        &gatt_char_value[CHARACTERISTIC_ONE], 20, 20,
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_WRITE |
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_NOTIFY);

    // create our LED characteristic (read, write)
    gatt_characteristics[CHARACTERISTIC_TWO] =
        new GattCharacteristic(
        nRF51_GATT_CHAR_TWO,
        &gatt_char_value[CHARACTERISTIC_TWO], 20, 500,
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_NOTIFY |
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_WRITE_WITHOUT_RESPONSE);

    // create our LED characteristic (read, write)
    gatt_characteristics[CHARACTERISTIC_THREE] =
        new GattCharacteristic(
        nRF51_GATT_CHAR_THREE,
        &gatt_char_value[CHARACTERISTIC_THREE], 20, 500,
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_NOTIFY |
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_WRITE_WITHOUT_RESPONSE);


    // create our LED characteristic (read, write)
    gatt_characteristics[CHARACTERISTIC_FOUR] =
        new GattCharacteristic(
        nRF51_GATT_CHAR_FOUR,
        &gatt_char_value[CHARACTERISTIC_FOUR], 20, 20,
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_WRITE |
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_NOTIFY);

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    // Display Model Number
    static char         model[7] = {'1','1','1','1','1','1','\0'};
    gattInfo_characteristics[CHARACTERISTIC_ONE] =
        new GattCharacteristic(
        GattCharacteristic::UUID_MODEL_NUMBER_STRING_CHAR,
        (uint8_t *)model, sizeof(model), sizeof(model),
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_READ);

    // Display Serial Number
    static char         serial[11] = {'2','2','2','2','2','2','2','2','2','2','\0'};
    gattInfo_characteristics[CHARACTERISTIC_TWO] =
        new GattCharacteristic(
        GattCharacteristic::UUID_SERIAL_NUMBER_STRING_CHAR,
        (uint8_t *)serial, sizeof(serial), sizeof(serial),
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_READ);

    // Display Harware Version
    static char         hwversion[9] = {'3','3','3','3','3','3','3','3','\0'};
    gattInfo_characteristics[CHARACTERISTIC_THREE] =
        new GattCharacteristic(
        GattCharacteristic::UUID_HARDWARE_REVISION_STRING_CHAR,
        (uint8_t *)hwversion, sizeof(hwversion), sizeof(hwversion),
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_READ);

    // Display Firmware Version
    static char         fwversion[10] = {'4','4','4','4','4','4','4','4','4','\0'};
    gattInfo_characteristics[CHARACTERISTIC_FOUR] =
        new GattCharacteristic(
        GattCharacteristic::UUID_FIRMWARE_REVISION_STRING_CHAR,
        (uint8_t *)fwversion, sizeof(fwversion), sizeof(fwversion),
        GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_READ);

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    // create our services with their respective characteristics
    gatt_service = new GattService(nRF51_GATT_SERVICE, gatt_characteristics, CHARACTERISTIC_COUNT);
    gatt_serviceONE = new GattService(GattService::UUID_DEVICE_INFORMATION_SERVICE, gattInfo_characteristics, CHARACTERISTIC_COUNT);


    // initialize our ble device
    ble.init();
    ble.setDeviceName((uint8_t *)DEVICE_NAME);
    ble.gap().setAddress(BLEProtocol::AddressType::RANDOM_STATIC, address1);

    // configure our advertising type, payload and interval
    ble.accumulateAdvertisingPayload(
        GapAdvertisingData::BREDR_NOT_SUPPORTED |
        GapAdvertisingData::LE_GENERAL_DISCOVERABLE);
    ble.accumulateAdvertisingPayload(
        GapAdvertisingData::COMPLETE_LOCAL_NAME,
        (uint8_t *)DEVICE_NAME, sizeof(DEVICE_NAME));
    ble.setAdvertisingType(GapAdvertisingParams::ADV_CONNECTABLE_UNDIRECTED);
    ble.setAdvertisingInterval(160); // 100ms

   

    ble.getPreferredConnectionParams(&fast);
    fast.minConnectionInterval = timerONE;
    fast.maxConnectionInterval = timerONE;
    fast.slaveLatency = 0;
    fast.connectionSupervisionTimeout = timerONE;
    ble.setPreferredConnectionParams(&fast);
    
     // configure our callbacks
    ble.onDisconnection(onDisconnection);
    ble.onConnection(onConnection);
    ble.onDataWritten(onDataWritten);

    // add our gatt service with two characteristics
    ble.addService(*gatt_service);
    ble.addService(*gatt_serviceONE);

    // start advertising
    ble.startAdvertising();

    //waiting for events
    for (;;) {
        ble.waitForEvent();
    }
}